//
//   Nintendo DS Save data Reader/Writer   by kako
//
//    version 1.2 - 1.4   modified for  Jubai's BBME firmware
//

#include <stdio.h>
#include <windows.h>
#include "cusb.h"
#include "fw_data.h"

#define VER "1.4"

#define RGIO_CPIPE 0 //R}hppCv(1OUT)ւ̔ԍ 
#define RGIO_RPIPE 7 //[hf[^pCv(1IN)ւ̔ԍ 

#define SFR_WRITE 0x00 
#define SFR_READ 0x01 
#define EZR_WRITE 0x02 
#define EZR_READ 0x03 
#define	EEP_WRITE	0x04  //BBMEFW̊g
#define	EEP_READ	0x05  //BBMEFW̊g
#define	EEP_READ_END	0x06  //BBMEFW̊g
#define	EEP_FORMAT	0x09  //BBMEFW̊g

#define _PORTACFG (u8)0x7F93; 
#define _PORTBCFG (u8)0x7F94; 
#define _PORTCCFG (u8)0x7F95; 
#define _OUTA (u8)0x7F96; 
#define _OUTB (u8)0x7F97; 
#define _OUTC (u8)0x7F98; 
#define _PINSA (u8)0x7F99; 
#define _PINSB (u8)0x7F9A; 
#define _PINSC (u8)0x7F9B; 
#define _OEA (u8)0x7F9C; 
#define _OEB (u8)0x7F9D; 
#define _OEC (u8)0x7F9E; 

FILE *fp;
unsigned char filename[512],pathname[512],filename2[512];
unsigned int indata[18];
unsigned char d0,d1,d2,d3,d4,d5,d6,d7,d8,d9,d10,d11,d12,d13,d14,d15;

HANDLE dev_handle;
unsigned char usb_buf[64];
unsigned char res_buf[64];
unsigned char lbuf[4096];
int buf_len;
#define MAXMEM 524288
unsigned char membuf[MAXMEM];
unsigned char membuf2[MAXMEM];
unsigned char header_a1[16];
unsigned char header_a2[16];
unsigned char header_b1[16];
unsigned char header_b2[16];
unsigned char header_c1[16];
unsigned char header_c2[16];

void progress(int n,int total)
{
	if (total<256) { return; }
	if (n%128==0) {
		printf("(%4.1f%%)\r", ((float)n*100)/total);
	}
}

void reset_buf(void)
{
	buf_len=0;
}

void flush_buf(void)
{
	if (buf_len>0) {
		usb_bulk_write(&dev_handle,RGIO_CPIPE,usb_buf,buf_len); 
		buf_len=0;
	}
}

void set_port_b_mode(int mode)
{
	usb_buf[buf_len++]=EZR_WRITE; 
	usb_buf[buf_len++]=_PORTBCFG; 
	usb_buf[buf_len++]=0x00; // All I/O port
	usb_buf[buf_len++]=EZR_WRITE; 
	usb_buf[buf_len++]=_OEB; 
	if (mode==0) {
		usb_buf[buf_len++]=0x00; // All Input
	} else {
		usb_buf[buf_len++]=0xff; // All Output
	}
}

void set_port_c_mode(int mode)
{
	usb_buf[buf_len++]=EZR_WRITE; 
	usb_buf[buf_len++]=_PORTCCFG; 
	usb_buf[buf_len++]=0x00; // All I/O port
	usb_buf[buf_len++]=EZR_WRITE; 
	usb_buf[buf_len++]=_OEC; 
	if (mode==0) {
		usb_buf[buf_len++]=0x00; // All Input
	} else {
		usb_buf[buf_len++]=0xff; // All Output
	}
}

void port_b_output(unsigned char d)
{
	usb_buf[buf_len++]=EZR_WRITE; //PORT-B̏o̓R}h 
	usb_buf[buf_len++]=_OUTB; 
	usb_buf[buf_len++]=d;
}

void port_c_output(unsigned char d)
{
	usb_buf[buf_len++]=EZR_WRITE; //PORT-C̏o̓R}h 
	usb_buf[buf_len++]=_OUTC; 
	usb_buf[buf_len++]=d;
}

void port_b_input(void)
{
	usb_buf[buf_len++]=EZR_READ; //PORT-B̓ǂݍ݃R}h 
	usb_buf[buf_len++]=_PINSB; 
}

void port_c_input(void)
{
	usb_buf[buf_len++]=EZR_READ; //PORT-C̓ǂݍ݃R}h 
	usb_buf[buf_len++]=_PINSC; 
}

void get_port_io_result(int len)
{
	int i;
	for (i=0;i<len;i++) { res_buf[i]=0x00; }
	usb_bulk_read(&dev_handle,RGIO_RPIPE,res_buf,len);
}

void eeprom_pinout(int a,int b,int c)
{
	unsigned char d;
	d=0x00;
	if (a>0) { d|= 0x04; } // CS
	if (b>0) { d|= 0x02; } // Serial-Clk
	if (c>0) { d|= 0x01; } // Data-Input
	port_b_output(d);
}

void CS_Low(void)
{
	eeprom_pinout(0,0,0);  //  CS,Clk,Dat
	flush_buf();
}

void CS_High(void)
{
	eeprom_pinout(1,0,0);  //  CS,Clk,Dat
	flush_buf();
}

void eeprom_send_bits(unsigned int data,int len)
{
	int i;
	for (i=0;i<len;i++) {
		data<<=1;
		if ( data & (1<<len) ) {
			eeprom_pinout(0,0,1);  //  CS,Clk,Dat
			eeprom_pinout(0,1,1);  //  CS,Clk,Dat
		} else {
			eeprom_pinout(0,0,0);  //  CS,Clk,Dat
			eeprom_pinout(0,1,0);  //  CS,Clk,Dat
		}
		if (i%8==7) { flush_buf(); }
	}
	eeprom_pinout(0,0,0);  //  CS,Clk,Dat
	flush_buf();
}

unsigned int  eeprom_receive_bits(int len)
{
	int i;
	unsigned int data;
	unsigned char a;
	for (i=0;i<len;i++) {
		eeprom_pinout(0,1,0);  //  CS,Clk,Dat
		port_c_input();
		eeprom_pinout(0,0,0);  //  CS,Clk,Dat
	}
	flush_buf();
	get_port_io_result(len);
	data=0;
	for (i=0;i<len;i++) {
		data<<=1;
		a=res_buf[i];
		if (a&0x01) { data|=0x01; }
	}
	return data;
}

void eeprom_read_address(int type,int ad)
{
	unsigned int data;
	unsigned char d;
	switch (type) {
	  case 1:
		// Send "READ" command
		if (ad<256) {
			data = 0x03; //  0000-0011 ... cmd 
		} else {
			data = 0x03 + 0x08; //  0000-1011 ... cmd 
		}
		eeprom_send_bits(data,8);
		data = ad%256; // xxxx-xxxx-xxxx-xxxx ... adrs
		eeprom_send_bits(data,8);
		break;
	  case 2:
		data = 0x03; //  0000-0011 ... cmd 
		eeprom_send_bits(data,8);
		data = ad; // xxxx-xxxx-xxxx-xxxx ... adrs
		eeprom_send_bits(data,16);
		break;
	  case 3:
		data = 0x03; //  0000-0011 ... cmd 
		eeprom_send_bits(data,8);
		data = ad; // xxxx-xxxx-xxxx-xxxx ... adrs
		eeprom_send_bits(data,24);
		break;
	}
}

unsigned char eeprom_read(int type,int ad)
{
	unsigned char d;
	unsigned int data;
	CS_Low();
	eeprom_read_address(type,ad);
	d=eeprom_receive_bits(8);
	CS_High();
	return d;
}

void eeprom_read_bytes(int type,int ad,unsigned char *buf,int size)
{
	unsigned char d;
	unsigned int data,i;
	CS_Low();
	eeprom_read_address(type,ad);
	for (i=0;i<size;i++) {
		progress(i,size);
		d=eeprom_receive_bits(8); buf[ad+i]=d;
	}
	CS_High();
}

void eeprom_write_address(int type,int ad)
{
	unsigned int data;
	unsigned char d;
	switch (type) {
	  case 1:
		// Send "WRITE" command
		if (ad<256) {
			data = 0x02; //  0000-0010 ... cmd 
		} else {
			data = 0x02 + 0x08; //  0000-1010 ... cmd 
		}
		eeprom_send_bits(data,8);
		data = ad%256; // xxxx-xxxx-xxxx-xxxx ... adrs
		eeprom_send_bits(data,8);
		break;
	  case 2:
		data = 0x02; //  0000-0010 ... cmd 
		eeprom_send_bits(data,8);
		data = ad; // xxxx-xxxx-xxxx-xxxx ... adrs
		eeprom_send_bits(data,16);
		break;
	  case 3:
		data = 0x02; //  0000-0010 ... cmd 
		eeprom_send_bits(data,8);
		data = ad; // xxxx-xxxx-xxxx-xxxx ... adrs
		eeprom_send_bits(data,24);
		break;
	}
}

//void eeprom_write_bytes(int type,int ad,unsigned char *buf,int size)
//{
//	int i;
//	unsigned int data;
//	CS_Low();
//	eeprom_write_address(type,ad);
//	for (i=0;i<size;i++) {
//		data=buf[i]; eeprom_send_bits(data,8);
//	}
//	CS_High();
//}

void eeprom_write(int type,int ad,unsigned char d)
{
	unsigned int data;
	CS_Low();
	eeprom_write_address(type,ad);
	data=d;
	eeprom_send_bits(data,8);
	CS_High();
}

void eeprom_write_enable(void)
{
	unsigned int data;
	CS_Low();
	// Send "WREN" command
	data = 0x06; //  0000-0110 ... cmd 
	eeprom_send_bits(data,8);
	CS_High();
}

unsigned int flash_id(void)
{
	unsigned int data;
	unsigned int id;
	unsigned char d;
	CS_Low();
	// Send "RDID" command
	data = 0x9F; //
	eeprom_send_bits(data,8);
	d=eeprom_receive_bits(8); id =  (d<<16);
	d=eeprom_receive_bits(8); id |= (d<<8);
	d=eeprom_receive_bits(8); id |= d;
	CS_High();
	return id;
}

void flash_sector_erase(int max_sector)
{
	unsigned int data;
	unsigned char d;
	int sector;
	CS_Low();
	printf("tbV̏Jn܂\n");
	
	for (sector=0;sector<max_sector;sector++) {
	
		// Send "WREN" command
		data = 0x06; //
		eeprom_send_bits(data,8);
		CS_High();
		
		printf("Sector %d erase ",sector);
		CS_Low();
		// Send "SE" command  ... sector erase
		data = 0xD8; //
		eeprom_send_bits(data,8);
		data = sector * 0x10000; // xxxx-xxxx-xxxx-xxxx ... adrs
		eeprom_send_bits(data,24);
		CS_High();
		
		CS_Low();
		// Send "RDSR" command
		data = 0x05; // Read Status Reg.
		eeprom_send_bits(data,8);
		while (1) {
			d=eeprom_receive_bits(8);
			if (d & 0x01) { printf("."); fflush(stdout); Sleep(300); continue; }
			printf("\n");
			break;
		}
		CS_High();
	}
}

void flash_write(unsigned char *buf,unsigned int ad,int size)
{
	unsigned int data;
	unsigned char d;
	int i;
	
	//printf("ad=%06X\n",ad);
	CS_Low();
	// Send "WREN" command
	data = 0x06; //
	eeprom_send_bits(data,8);
	CS_High();
	
	CS_Low();
	// Send "PP" command  ... Page program
	data = 0x02; //
	eeprom_send_bits(data,8);
	data = ad; // xxxx-xxxx-xxxx-xxxx ... adrs
	eeprom_send_bits(data,24);
	for (i=0;i<256;i++) {
		data=buf[ad+i];
		eeprom_send_bits(data,8);
	}
	CS_High();

	CS_Low();
	// Send "RDSR" command
	data = 0x05; // Read Status Reg.
	eeprom_send_bits(data,8);
	while (1) {
		d=eeprom_receive_bits(8);
		if (d & 0x01) { Sleep(100); continue; }
		break;
	}
	CS_High();
}

int header_cmp(unsigned char *buf1,unsigned char *buf2)
{
	int i;
	for (i=0;i<16;i++) {
		if (buf1[i]!=buf2[i]) { return -1; }
	}
	return 0;
}

void print_header(unsigned char *buf)
{
	int i;
	printf(">>>>");
	for (i=0;i<16;i++) { printf(" %02X",buf[i]); }
	printf("\n");
}

void RomReadBytes(u32 address,u32 size,char dump){
	u8 i;
	u32 t,l;
	//Rom̓ewAhXǂݏoB
	//ʂrom_buf[]Ɏ[B߂l͖łB

		i=0;
		usb_buf[i++]=EEP_READ;//PORT-C̓ǂݍ݃R}h
		usb_buf[i++]=_OUTB;
		usb_buf[i++]=_PINSC;
		usb_buf[i++]=address / 0x10000;//23-16
		usb_buf[i++]=address / 0x100; //15-08
		usb_buf[i++]=address;//07-00
		usb_buf[i++]=1;//Aǂݍ݃tO
		usb_bulk_write(&dev_handle,RGIO_CPIPE,usb_buf,i);

	//ǂݍ݂̃[vB
	l=0;
	while(l < size){
		usb_bulk_read(&dev_handle,RGIO_RPIPE,usb_buf,64);
		t=0;
		while(t < 64 && (l+t) < size){
			membuf[l+t] = usb_buf[t]; //ROMPoCgǂݍ

			if(dump == 1){
				if((t % 16) == 0){
					printf("\n%2.6X ",l+t+address);
				}
				printf("%2.2X ",membuf[l+t]);
			}
			else if(dump == 2){
				printf("READ : %2.6X\r ",l+t+address);
			}
			else if(dump == 3){
				progress(l,size);
			}
			t++;
		}
	l=l+64;
	}

	//ǂݍ݃ZbVI
	i=0;
	usb_buf[i++]=EEP_READ_END;//PORT-C̓ǂݍ݃R}h
	usb_bulk_write(&dev_handle,RGIO_CPIPE,usb_buf,i);
	usb_bulk_read(&dev_handle,RGIO_RPIPE,usb_buf,64);//]obt@ǂݍ
}

void RomWriteBytes(u32 address,u32 size,u32 file_address,u8 dump){
	u16 i;u32 t;
	//ROM̓eB

	i=0;
	usb_buf[i++]=EEP_WRITE;//PORT-C̓ǂݍ݃R}h
	usb_buf[i++]=_OUTB;
	usb_buf[i++]=_PINSC;
	usb_buf[i++]=address / 0x10000;//23-16
	usb_buf[i++]=address / 0x100; //15-08
	usb_buf[i++]=address;//07-00
	usb_buf[i++]=size / 0x10000;//23-16
	usb_buf[i++]=size / 0x100; //15-08
	usb_buf[i++]=size;//07-00
	usb_bulk_write(&dev_handle,RGIO_CPIPE,usb_buf,i);

	//݂̃[vB
	t=0;
	while(t < size){
		i=0;
		while(i < 64 && t < size){
			usb_buf[i] = membuf2[t + file_address];
			i++;t++;
		}
		usb_bulk_write(&dev_handle,RGIO_CPIPE,usb_buf,i);

		if((dump == 2)){
			printf("WRITE : %2.6X \r ",t+address);
		}
		else if(dump == 3){
			progress(t,size);
		}
	}
}

void more_()
{
	printf("[ L[ĉ ]\n"); getch();
}

int main(int argc,char **argv)
{
	int ad,ad0,i,n,header_type,save_size,hex_size,mode_rw,l,m,file_type,max_flash_sectors;
	unsigned int id;
	unsigned char d;

	printf("<<< NintendoDS SaveData Reader/Writer Ver%s by E.Kako >>>\n",VER);
	printf("\n");
	
	mode_rw=0; // read mode
	file_type=0;
	if ((argc!=1)&&(argc!=2)) {
		printf("Usage: ndswrt       ... read\n");
		printf("       ndswrt file  ... write\n");
		printf("       ndswrt -e    ... erase\n");
		more_(); exit(1);
	}
	// Ƃăt@Cw肳ꂽꍇÃt@Cǂݍ
	if (argc==2) {
		mode_rw=1; // write
		if ( (strcmp(argv[1],"-e")==0)||(strcmp(argv[1],"-E")==0) ) {
			mode_rw=2; // erase
		} else {
			strcpy(filename,argv[1]);
			l=strlen(filename); m= -1;
			for (i=0;i<l;i++) {
				if (filename[i]=='\\') { m=i; }
			}
			if (m>0) {
				strcpy(pathname,filename); 
				pathname[m+1]='\0';
				if (strcmp(filename+m+1,"backup.txt")==0) {
					printf("Error: t@Cbackup.txt̓l[ĂgĂ\n");
					more_(); exit(1);
				}
				if (strcmp(filename+m+1,"backup.sav")==0) {
					printf("Error: t@Cbackup.sav̓l[ĂgĂ\n");
					more_(); exit(1);
				}
			} else {
				pathname[0]='\0';
				if (strcmp(filename,"backup.txt")==0) {
					printf("Error: t@Cbackup.txt̓l[ĂgĂ\n");
					more_(); exit(1);
				}
				if (strcmp(filename,"backup.sav")==0) {
					printf("Error: t@Cbackup.sav̓l[ĂgĂ\n");
					more_(); exit(1);
				}
			}
		}
	}
	if (mode_rw==1) {
		l=strlen(filename);
		if ( (filename[l-4]=='.')&&(filename[l-3]=='t')&&(filename[l-2]=='x')&&(filename[l-1]=='t')) {
			file_type=1; // .txt    textfile
		}
		if ( (filename[l-4]=='.')&&(filename[l-3]=='s')&&(filename[l-2]=='a')&&(filename[l-1]=='v')) {
			file_type=2; // .sav    binaryfile
		}
		if (file_type==0) {
			fp=fopen(filename,"rb");
			hex_size= filelength( fileno(fp) );
			for (i=0;i<16;i++) { membuf2[i]=0x00; }
			fread( membuf2 , 1 , 16 , fp );
			fclose(fp);
			l= -1;
			for (i=0;i<16;i++) {
				if ((membuf2[i]<' ')&&(membuf2[i]>0x7E)) { l=i; } // detect non-ascii-code
			}
			if (l== -1) { // ascii code
				file_type=1;
			} else {
				file_type=2; // maybe binary
			}
		}
		
		if (file_type==1) {
			printf("eLXgt@C(%s)ǂݍ݂܂\n",filename);
			fp=fopen(filename,"r");
			if (fp==NULL) {
				printf("Error: t@C܂(%s)\n",filename);
				more_(); exit(1);
			}
			for (i=0;i<MAXMEM;i++) { membuf2[i]=0xFF; }
			hex_size=0;
			while (1) {
				if (fgets(lbuf,4096,fp)==NULL) { break; }
				if (lbuf[0]=='\n') { continue; }
				//n=sscanf(lbuf,"%X %X %X %X %X %X %X %X %X %X %X %X %X %X %X %X %X",
				// &ad,&d0,&d1,&d2,&d3,&d4,&d5,&d6,&d7,&d8,&d9,&d10,&d11,&d12,&d13,&d14,&d15);
				l=0; n=0;
				for (i=0;i<strlen(lbuf);i++) {
					if ((lbuf[i]==' ')||(lbuf[i]=='\n')) {
						sscanf(lbuf+l,"%X",&indata[n]); l=i+1; n++;
					}
				}
				if (n!=17) {
					printf("Error: t@Č`Ⴂ܂(%d)\n",n); fclose(fp); more_(); exit(1);
				}
				ad = indata[0];
				d0 = indata[1];  d1 = indata[2];  d2 = indata[3];  d3 = indata[4];
				d4 = indata[5];  d5 = indata[6];  d6 = indata[7];  d7 = indata[8];
				d8 = indata[9];  d9 = indata[10];  d10 = indata[11];  d11 = indata[12];
				d12 = indata[13];  d13 = indata[14];  d14 = indata[15];  d15 = indata[16];
				if (ad>MAXMEM) {
					printf("Error: AhX傫߂ł(%d)\n",ad); fclose(fp); more_(); exit(1);
				}
				membuf2[ ad + 0 ] = d0; membuf2[ ad + 1 ] = d1; membuf2[ ad + 2 ] = d2;
				membuf2[ ad + 3 ] = d3; membuf2[ ad + 4 ] = d4; membuf2[ ad + 5 ] = d5;
				membuf2[ ad + 6 ] = d6; membuf2[ ad + 7 ] = d7; membuf2[ ad + 8 ] = d8;
				membuf2[ ad + 9 ] = d9; membuf2[ ad + 10 ] = d10; membuf2[ ad + 11 ] = d11;
				membuf2[ ad + 12 ] = d12; membuf2[ ad + 13 ] = d13; membuf2[ ad + 14 ] = d14;
				membuf2[ ad + 15 ] = d15;
				if (ad+16>hex_size) { hex_size=ad+16; }
			}
			fclose(fp);
			
		} else if (file_type==2) {
			printf("oCit@C(%s)ǂݍ݂܂\n",filename);
			fp=fopen(filename,"rb");
			if (fp==NULL) {
				printf("Error: t@C܂(%s)\n",filename);
				more_(); exit(1);
			}
			hex_size= filelength( fileno(fp) );
			if (hex_size > MAXMEM) {
				fclose(fp);
				printf("Error: t@C傫߂ł(%d)\n",hex_size); fclose(fp); more_(); exit(1);
			}
			for (i=0;i<MAXMEM;i++) { membuf2[i]=0xFF; }
			fread( membuf2 , 1 , hex_size , fp );
			fclose(fp);
		}
		printf("t@C(%s)ǂݍ݂܂\n",filename);
	}
	
	// t@[EFȀ
	if(cusb_init(0,&dev_handle,fw_bin,"RGIO","V100")){
		printf("Error: EZ-USB܂ł\n"); more_(); exit(0);
	}
	
	// |[g̏
	reset_buf();
	set_port_b_mode(1); // PortB ... Output
	set_port_c_mode(0); // PortC ... Input
	eeprom_pinout(1,0,0);  //  CS,Clk,Dat
	flush_buf();
	
	printf("ID`FbN܂\n");
	id=flash_id(); // dummy read
	id=flash_id();
	printf("ID=%06X\n",id);
	
	header_type=0;
	if (id == 0xFFFFFF) {
		// tbVłȂ̂ŁAIDǂ݂ƂȂ
	} else if (id == 0x621600) {
		header_type = 3;  save_size = 262144;
		printf("VAFLASH SANYO LE25FW203 2Mbitł\n");
	} else if (id == 0x204012) {
		header_type = 3;  save_size = 262144;
		printf("VAFLASH STMicroelectronics M45PE20 2Mbitł\n");
	} else if (id == 0x204013) {
		header_type = 3;  save_size = 524288;
		printf("VAFLASH STMicroelectronics M45PE40 4Mbitł\n");
	} else {
		printf("m̃VAFLASHł\n");
	}
	
	if (header_type==0) {
		printf("Z[up̐擪ǂݍŁAZ[u𔻕ʂ܂\n");
		// Z[up̃wb_̃`FbN
		eeprom_read_bytes(1,0,header_a1,16);
		for (ad=0;ad<16;ad++) { header_a2[ad]=eeprom_read(1,ad); }
		eeprom_read_bytes(2,0,header_b1,16);
		for (ad=0;ad<16;ad++) { header_b2[ad]=eeprom_read(2,ad); }
		eeprom_read_bytes(3,0,header_c1,16);
		for (ad=0;ad<16;ad++) { header_c2[ad]=eeprom_read(3,ad); }
	
		// tH[}bg̃`FbNp
		for (ad=0;ad<16;ad++) { membuf[ad]=0xFF; }
		
		if ((header_cmp(header_a1,header_a2)==0)&&(header_cmp(header_a1,membuf)!=0)) {
			header_type = 1;  save_size = 512;
			printf("VAEEPROM STMicroelectronics M95040 4kbitł\n");
		} else if ((header_cmp(header_b1,header_b2)==0)&&(header_cmp(header_b1,membuf)!=0)) {
			header_type = 2;  //save_size = 8192;
			//printf("VAEEPROM STMicroelectronics M95640 64kbitł\n");
			// ܂肵Ȃ
		} else if ((header_cmp(header_c1,header_c2)==0)&&(header_cmp(header_c1,membuf)!=0)) {
			header_type = 3;  save_size = 262144;
			printf("VAFLASH SANYO LE25FW203 2Mbitł\n");
		}
	}
	
	if (header_type==0) {
		// COŃbW[T[A炫Gh́A
		// EEPROM̐擪 0xFFŖ܂Ă̂ŃŜ`FbN
		// (邢͑ɂ̂邩Ȃ?)
		RomReadBytes(0,16384,0); // 8kbyte*2 ǂݍ (1oCgēǂݍ܂邪CɂȂ)
		for (ad=0;ad<16384;ad++) {
			if (membuf[ad]!=0xFF) { ad0=ad; header_type= -1; break; }
		}
		if (header_type==0) {
			printf("̃J[gbW͂炭tH[}bgł\n");
			printf("`ʂłȂ߁AǂݏłȂȂ\n");
			printf("(邢̓J[gbWƑ}Ă܂)\n");
		}
	}
	
	if (header_type== -1) {
		if (header_cmp(membuf+ad0,membuf+ad0+8192)==0) {
			printf("VAEEPROM STMicroelectronics M95640 64kbitł\n");
			header_type=2; save_size=8192;
		} else {
			printf("VAEEPROM STMicroelectronics M95512 512kbitł\n");
			save_size = 65536;
		}
	} else if (header_type==2) {
		RomReadBytes(0,16384,0); // 8kbyte*2 ǂݍ (1oCgēǂݍ܂邪CɂȂ)
		if (header_cmp(membuf,membuf+8192)==0) {
			printf("VAEEPROM STMicroelectronics M95640 64kbitł\n");
			save_size = 8192;
		} else {
			printf("VAEEPROM STMicroelectronics M95512 512kbitł\n");
			save_size = 65536;
		}
	}
	
	if (header_type==0) {
		if (mode_rw==1) {
			if (hex_size==512) {
				header_type = 1;  save_size = 512;
				printf("VAEEPROM STMicroelectronics M95040 4kbitƉ\n");
			} else if (hex_size==8192) {
				header_type = 2;  save_size = 8192;
				printf("VAEEPROM STMicroelectronics M95640 64kbitƉ\n");
				} else if (hex_size==262144) {
				header_type = 3;  save_size = 262144;
				printf("VAFLASH SANYO LE25FW203 2MbitƉ\n");
			}
		}
	}
	
	if (header_type==3) {
		max_flash_sectors = save_size/65536;
	}
	
	
	if (header_type==0) {
		printf("m̕Ȃ̂ŁAł܂ł\n");
		usb_close(&dev_handle);
		more_(); exit(1);
	}
	
	if (mode_rw==0) {
		printf("J[gbWf[^̓ǂݍݒł...\n");
		if ((header_type==1)||(header_type==2)) {
			eeprom_read_bytes(header_type,0,membuf,save_size);
		} else {
			RomReadBytes(0,save_size,3); // BBME̊֐𗬗p
		}
		//t@Cɕۑ
		fp=fopen("savedata.txt","w");
		for (ad=0;ad<save_size;ad++) {
			if (ad%16==0) { fprintf(fp,"%04X",ad); }
			d=membuf[ad];
			fprintf(fp," %02X",d);
			if (ad%16==15) { fprintf(fp,"\n"); }
		}
		fprintf(fp,"\n");
		fclose(fp);
		fp=fopen("savedata.sav","wb");
		for (ad=0;ad<save_size;ad++) {
			d=membuf[ad]; fprintf(fp,"%c",d);
		}
		fclose(fp);
		printf("ۑ܂(savedata.txt, savedata.sav)\n");
		
	} else if (mode_rw==1) {
		
		if (save_size != hex_size) {
			printf("Error: J[gbWƏ݂t@C̃TCYقȂ\n");
			printf("(cart=%d,file=%d)\n",save_size,hex_size);
			usb_close(&dev_handle);
			more_(); exit(1);
		}
		
		printf("J[gbWɏ݂OɃobNAbv܂\n");
		
		printf("J[gbWf[^ǂݍݒł...\n");
		if ((header_type==1)||(header_type==2)) {
			eeprom_read_bytes(header_type,0,membuf,save_size);
		} else {
			RomReadBytes(0,save_size,3); // BBME̊֐𗬗p
		}
	
		//t@Cɕۑ
		sprintf(filename2,"%s%s",pathname,"backup.txt");
		fp=fopen(filename2,"w");
		for (ad=0;ad<save_size;ad++) {
			if (ad%16==0) { fprintf(fp,"%04X",ad); }
			d=membuf[ad];
			fprintf(fp," %02X",d);
			if (ad%16==15) { fprintf(fp,"\n"); }
		}
		fprintf(fp,"\n");
		fclose(fp);
		sprintf(filename2,"%s%s",pathname,"backup.sav");
		fp=fopen(filename2,"wb");
		for (ad=0;ad<save_size;ad++) {
			d=membuf[ad]; fprintf(fp,"%c",d);
		}
		fclose(fp);
		printf("obNAbv܂(backup.txt, backup.sav)\n");
		
		if (header_type==3) {
			printf("ݑOɁAtbV̏s܂\n");
			printf("J[gbWf[^ł...\n");
			flash_sector_erase(max_flash_sectors);
			
			printf("f[^A݂Jn܂\n");
			//for (ad=0;ad<save_size;ad+=256) {
			//	progress(ad,save_size);
			//	flash_write(membuf2,ad,256);
			//}
			RomWriteBytes(0,save_size,0,3); // BBME̊֐𗬗p
			
			Sleep(3000);
			printf("݊\n");
			
		} else {
		
			printf("J[gbWf[^ݒł...\n");
			for (ad=0;ad<save_size;ad++) {
				progress(ad,save_size);
				eeprom_write_enable();
				eeprom_write(header_type,ad,membuf2[ad]);
			}
			Sleep(3000);
		}
		
		printf("݌ʂxt@C܂\n");
		
		printf("J[gbWf[^ǂݍݒł...\n");
		if ((header_type==1)||(header_type==2)) {
			eeprom_read_bytes(header_type,0,membuf,save_size);
		} else {
			RomReadBytes(0,save_size,3); // BBME̊֐𗬗p
		}
		
		for (ad=0;ad<save_size;ad++) {
			//printf("%04X %02X %02X\n",ad,membuf[ad],membuf2[ad]);
			if ( membuf[ad] != membuf2[ad] ) {
				printf("Error: ܂Ă܂(adr=%04X)\n",ad);
				printf("(xA݂݂ĉ)\n");
				usb_close(&dev_handle);
				more_(); exit(1);
			}
		}
		printf("܂Ă܂\n");
		
	} else if (mode_rw==2) {
		
		printf("J[gbWf[^OɃobNAbv܂\n");
		
		printf("J[gbWf[^ǂݍݒł...\n");
		if ((header_type==1)||(header_type==2)) {
			eeprom_read_bytes(header_type,0,membuf,save_size);
		} else {
			RomReadBytes(0,save_size,3); // BBME̊֐𗬗p
		}
	
		//t@Cɕۑ
		fp=fopen("backup.txt","w");
		for (ad=0;ad<save_size;ad++) {
			if (ad%16==0) { fprintf(fp,"%04X",ad); }
			d=membuf[ad];
			fprintf(fp," %02X",d);
			if (ad%16==15) { fprintf(fp,"\n"); }
		}
		fprintf(fp,"\n");
		fclose(fp);
		fp=fopen("backup.sav","wb");
		for (ad=0;ad<save_size;ad++) {
			d=membuf[ad]; fprintf(fp,"%c",d);
		}
		fclose(fp);
		printf("obNAbv܂(backup.txt, backup.sav)\n");
		
		if (header_type==3) {
			printf("J[gbWf[^ł...\n");
			flash_sector_erase(max_flash_sectors);
			
		} else {
			printf("J[gbWf[^ł...\n");
			for (ad=0;ad<save_size;ad++) {
				progress(ad,save_size);
				eeprom_write_enable();
				eeprom_write(header_type,ad,0xFF);
			}
			Sleep(1000);
		}
	}

	//I
	set_port_b_mode(0); // PortB ... Input
	set_port_c_mode(0); // PortC ... Input
	usb_close(&dev_handle);
	
	printf("I܂B\n"); more_();
	return 0;
}
